"""Comprehensive tests for arklex.env.tools.tools (Tool class and register_tool)."""

from typing import NoReturn
from unittest.mock import Mock

# Mocks for dependencies
from arklex.orchestrator.entities.orchestrator_state_entities import (
    OrchestratorState,
    StatusEnum,
)
from arklex.orchestrator.nlu.entities.slot_entities import Slot
from arklex.resources.tools import tools
from arklex.resources.tools.tools import Tool, register_tool


def dummy_func(a: object = None, b: object = None, **kwargs: object) -> str:
    return f"a={a}, b={b}"


def test_register_tool_decorator_creates_tool() -> None:
    """Test that register_tool decorates a function and returns a Tool."""
    tool_instance = register_tool(
        description="desc", slots=[{"name": "a", "type": "str", "description": "A"}]
    )(dummy_func)
    assert isinstance(tool_instance, Tool)
    assert tool_instance.description == "desc"
    assert any(slot.name == "a" for slot in tool_instance.slots)


def test_tool_init_slotfiller() -> None:
    """Test Tool.init_slotfiller sets the slotfiller attribute."""
    tool = Tool(dummy_func, "toolname", "desc", [])
    mock_sf = Mock()
    tool.init_slotfiller(mock_sf)
    assert tool.slotfiller is mock_sf


def test_tool__init_slots_populates_slots() -> None:
    """Test _init_slots populates slots from state.default_slots."""
    slot = Slot(
        name="a", type="str", description="A", required=True, value="foo", verified=True
    )
    state = Mock(spec=OrchestratorState)
    default_slots = {"default_slots": [slot]}
    state.function_calling_trajectory = []
    tool = Tool(
        dummy_func,
        "toolname",
        "desc",
        [{"name": "a", "type": "str", "description": "A", "required": True}],
    )
    tool._init_slots(state, default_slots)
    assert tool.slots[0].value == "foo"
    assert tool.slots[0].verified is True
    assert state.function_calling_trajectory


def test_tool_execute_successful() -> None:
    """Test Tool.execute with all slots filled and verified."""
    slot = Slot(
        name="a", type="str", description="A", required=True, value="bar", verified=True
    )
    state = Mock(spec=OrchestratorState)
    state.slots = {}
    state.function_calling_trajectory = []
    state.trajectory = [[Mock(input=None, output=None)]]
    state.bot_config = Mock()
    state.bot_config.llm_config = Mock()
    state.bot_config.llm_config.model_dump.return_value = {}
    state.message_flow = ""
    state.status = StatusEnum.INCOMPLETE
    tool = Tool(
        dummy_func,
        "toolname",
        "desc",
        [{"name": "a", "type": "str", "description": "A", "required": True}],
    )
    tool.slots = [slot]
    tool.slotfiller = Mock()
    tool.slotfiller.fill_slots.return_value = [slot]
    tool.isResponse = False
    result_state, result_output = tool.execute(state, {}, {})
    assert result_output.status == StatusEnum.COMPLETE
    assert result_output.slots["toolname"][0].value == "bar"


def test_tool_execute_incomplete_due_to_missing_slot() -> None:
    """Test Tool.execute when a required slot is missing."""
    slot = Slot(
        name="a",
        type="str",
        description="A",
        required=True,
        value=None,
        verified=False,
        prompt="Prompt for a",
    )
    state = Mock(spec=OrchestratorState)
    state.slots = {}
    state.function_calling_trajectory = []
    state.trajectory = [[Mock(input=None, output=None)]]
    state.bot_config = Mock()
    state.bot_config.llm_config = Mock()
    state.bot_config.llm_config.model_dump.return_value = {}
    state.message_flow = ""
    tool = Tool(
        dummy_func,
        "toolname",
        "desc",
        [{"name": "a", "type": "str", "description": "A", "required": True}],
    )
    tool.slots = [slot]
    tool.slotfiller = Mock()
    tool.slotfiller.fill_slots.return_value = [slot]
    result_state, result_output = tool.execute(state, {}, {})
    assert result_output.status == StatusEnum.INCOMPLETE
    assert (
        "Prompt for a" in result_state.message_flow or result_state.message_flow == ""
    )


def test_tool_execute_slot_verification_needed() -> None:
    """Test Tool.execute when slot verification is needed."""
    slot = Slot(
        name="a",
        type="str",
        description="A",
        required=True,
        value="foo",
        verified=False,
        prompt="Prompt for a",
    )
    state = Mock(spec=OrchestratorState)
    state.slots = {}
    state.function_calling_trajectory = []
    state.trajectory = [[Mock(input=None, output=None)]]
    state.bot_config = Mock()
    state.bot_config.llm_config = Mock()
    state.bot_config.llm_config.model_dump.return_value = {}
    state.message_flow = ""
    tool = Tool(
        dummy_func,
        "toolname",
        "desc",
        [{"name": "a", "type": "str", "description": "A", "required": True}],
    )
    tool.slots = [slot]
    tool.slotfiller = Mock()
    tool.slotfiller.fill_slots.return_value = [slot]
    tool.slotfiller.verify_slot.return_value = (True, "Need verification")
    result_state, result_output = tool.execute(state, {}, {})
    assert result_output.status == StatusEnum.INCOMPLETE


def test_tool_execute_tool_execution_error() -> None:
    """Test Tool.execute handles ToolExecutionError."""

    def error_func(**kwargs: object) -> NoReturn:
        raise tools.ToolExecutionError("toolname", "fail", extra_message="extra")

    slot = Slot(
        name="a", type="str", description="A", required=True, value="foo", verified=True
    )
    state = Mock(spec=OrchestratorState)
    state.slots = {}
    state.function_calling_trajectory = []
    state.trajectory = [[Mock(input=None, output=None)]]
    state.bot_config = Mock()
    state.bot_config.llm_config = Mock()
    state.bot_config.llm_config.model_dump.return_value = {}
    state.message_flow = ""
    tool = Tool(
        error_func,
        "toolname",
        "desc",
        [{"name": "a", "type": "str", "description": "A", "required": True}],
    )
    tool.slots = [slot]
    tool.slotfiller = Mock()
    tool.slotfiller.fill_slots.return_value = [slot]
    result_state, result_output = tool.execute(state, {}, {})
    assert result_output.status == StatusEnum.INCOMPLETE


def test_tool_execute_authentication_error() -> None:
    """Test Tool.execute handles AuthenticationError."""

    def error_func(**kwargs: object) -> NoReturn:
        raise tools.AuthenticationError("auth fail")

    slot = Slot(
        name="a", type="str", description="A", required=True, value="foo", verified=True
    )
    state = Mock(spec=OrchestratorState)
    state.slots = {}
    state.function_calling_trajectory = []
    state.trajectory = [[Mock(input=None, output=None)]]
    state.bot_config = Mock()
    state.bot_config.llm_config = Mock()
    state.bot_config.llm_config.model_dump.return_value = {}
    state.message_flow = ""
    tool = Tool(
        error_func,
        "toolname",
        "desc",
        [{"name": "a", "type": "str", "description": "A", "required": True}],
    )
    tool.slots = [slot]
    tool.slotfiller = Mock()
    tool.slotfiller.fill_slots.return_value = [slot]
    result_state, result_output = tool.execute(state, {}, {})
    assert result_output.status == StatusEnum.INCOMPLETE


def test_tool_str_and_repr() -> None:
    """Test __str__ and __repr__ methods."""
    tool = Tool(dummy_func, "toolname", "desc", [])
    assert str(tool) == "Tool"
    assert repr(tool) == "Tool"


# Tests for _validate_and_clean_slot_values ----------------------------------------------------------
def test_validate_and_clean_slot_values_simple_object_array() -> None:
    """Test validation of a simple array of objects."""
    slot_schema_object_array = {
        "type": "object",
        "required": ["vehicles_array"],
        "properties": {
            "vehicles_array": {
                "type": "array",
                "items": {
                    "type": "object",
                    "required": ["make", "model"],
                    "properties": {
                        "make": {
                            "type": "string",
                            "description": "make of the vehicle",
                        },
                        "model": {
                            "type": "string",
                            "description": "model of the vehicle",
                        },
                    },
                },
                "description": "Array of vehicle objects",
            }
        },
    }

    slot = Slot(
        name="vehicles_array",
        type="array",
        description="Array of vehicles",
        required=True,
        value=[{"make": "Toyota", "model": "Camry"}],
        verified=True,
        slot_schema={
            "type": "function",
            "function": {"name": "tool", "parameters": slot_schema_object_array},
        },
    )

    tool = Tool(dummy_func, "toolname", "desc", [])
    tool.slots = [slot]
    tool._validate_and_clean_slot_values()

    assert tool.slots[0].value == [{"make": "Toyota", "model": "Camry"}]


def test_validate_and_clean_slot_values_removes_extra_fields() -> None:
    """Test that validation removes extra fields not in schema."""
    slot_schema = {
        "type": "object",
        "required": ["vehicles_array"],
        "properties": {
            "vehicles_array": {
                "type": "array",
                "items": {
                    "type": "object",
                    "required": ["make", "model"],
                    "properties": {
                        "make": {"type": "string"},
                        "model": {"type": "string"},
                    },
                },
            }
        },
    }

    slot = Slot(
        name="vehicles_array",
        type="array",
        description="Array of vehicles",
        required=True,
        value=[
            {
                "make": "Toyota",
                "model": "Camry",
                "extra_field": "should_be_removed",
            }
        ],
        verified=True,
        slot_schema={
            "type": "function",
            "function": {"name": "tool", "parameters": slot_schema},
        },
    )

    tool = Tool(dummy_func, "toolname", "desc", [])
    tool.slots = [slot]
    tool._validate_and_clean_slot_values()

    # Extra field should be removed
    assert tool.slots[0].value == [{"make": "Toyota", "model": "Camry"}]
    assert "extra_field" not in tool.slots[0].value[0]


def test_validate_and_clean_slot_values_nested_objects() -> None:
    """Test validation of nested objects within array."""
    slot_schema_nested = {
        "type": "object",
        "required": ["cars_array"],
        "properties": {
            "cars_array": {
                "type": "array",
                "items": {
                    "type": "object",
                    "required": ["make", "model", "car"],
                    "properties": {
                        "car": {
                            "type": "object",
                            "required": ["car_id", "car_insurance_type"],
                            "properties": {
                                "car_id": {"type": "string"},
                                "car_insurance_type": {"type": "integer"},
                            },
                        },
                        "make": {"type": "string"},
                        "model": {"type": "string"},
                    },
                },
            }
        },
    }

    slot = Slot(
        name="cars_array",
        type="array",
        description="Array of cars",
        required=True,
        value=[
            {
                "make": "Toyota",
                "model": "Camry",
                "car": {
                    "car_id": "toyota_camry_2025",
                    "car_insurance_type": 1,
                    "dummy_field": "should_be_removed",
                },
            }
        ],
        verified=True,
        slot_schema={
            "type": "function",
            "function": {"name": "tool", "parameters": slot_schema_nested},
        },
    )

    tool = Tool(dummy_func, "toolname", "desc", [])
    tool.slots = [slot]
    tool._validate_and_clean_slot_values()

    # Check that dummy_field was removed from nested object
    validated_value = tool.slots[0].value[0]
    assert validated_value["make"] == "Toyota"
    assert validated_value["model"] == "Camry"
    assert validated_value["car"]["car_id"] == "toyota_camry_2025"
    assert validated_value["car"]["car_insurance_type"] == 1
    assert "dummy_field" not in validated_value["car"]


def test_validate_and_clean_slot_values_optional_fields() -> None:
    """Test validation with optional fields (not all required)."""
    object_schema = {
        "type": "object",
        "required": ["cars_array"],
        "properties": {
            "cars_array": {
                "type": "array",
                "items": {
                    "type": "object",
                    "required": ["make", "model", "car"],
                    "properties": {
                        "car": {
                            "type": "object",
                            "required": ["car_id"],
                            "properties": {
                                "car_id": {"type": "string"},
                                "car_insurance_type": {
                                    "type": "integer"
                                },  # Not required
                            },
                        },
                        "make": {"type": "string"},
                        "model": {"type": "string"},
                    },
                },
            }
        },
    }

    # Test with optional field present
    slot = Slot(
        name="cars_array",
        type="array",
        description="Array of cars",
        required=True,
        value=[
            {
                "make": "Toyota",
                "model": "Camry",
                "car": {"car_id": "toyota_camry_2025", "car_insurance_type": 1},
            }
        ],
        verified=True,
        slot_schema={
            "type": "function",
            "function": {"name": "tool", "parameters": object_schema},
        },
    )

    tool = Tool(dummy_func, "toolname", "desc", [])
    tool.slots = [slot]
    tool._validate_and_clean_slot_values()

    assert tool.slots[0].value[0]["car"]["car_insurance_type"] == 1

    # Test with optional field missing
    slot2 = Slot(
        name="cars_array",
        type="array",
        description="Array of cars",
        required=True,
        value=[
            {
                "make": "Honda",
                "model": "Accord",
                "car": {"car_id": "honda_accord_2025"},
            }
        ],
        verified=True,
        slot_schema={
            "type": "function",
            "function": {"name": "tool", "parameters": object_schema},
        },
    )

    tool2 = Tool(dummy_func, "toolname", "desc", [])
    tool2.slots = [slot2]
    tool2._validate_and_clean_slot_values()

    # Optional field should be set to None
    assert tool2.slots[0].value[0]["car"]["car_insurance_type"] is None


def test_validate_and_clean_slot_values_type_conversion() -> None:
    """Test that validation performs type conversion."""
    object_schema = {
        "type": "object",
        "required": ["data"],
        "properties": {
            "data": {
                "type": "object",
                "required": ["count", "price", "active"],
                "properties": {
                    "count": {"type": "integer"},
                    "price": {"type": "number"},
                    "active": {"type": "boolean"},
                },
            }
        },
    }

    slot = Slot(
        name="data",
        type="object",
        description="Test data",
        required=True,
        value={"count": 42, "price": 99.99, "active": True},
        verified=True,
        slot_schema={
            "type": "function",
            "function": {"name": "tool", "parameters": object_schema},
        },
    )

    tool = Tool(dummy_func, "toolname", "desc", [])
    tool.slots = [slot]
    tool._validate_and_clean_slot_values()

    validated_value = tool.slots[0].value
    assert isinstance(validated_value["count"], int)
    assert isinstance(validated_value["price"], float)
    assert isinstance(validated_value["active"], bool)


def test_validate_and_clean_slot_values_missing_required_field() -> None:
    """Test that validation raises error for missing required fields."""
    from arklex.utils.logging.exceptions import ValidationError

    object_schema = {
        "type": "object",
        "required": ["vehicles_array"],
        "properties": {
            "vehicles_array": {
                "type": "array",
                "items": {
                    "type": "object",
                    "required": ["make", "model"],
                    "properties": {
                        "make": {"type": "string"},
                        "model": {"type": "string"},
                    },
                },
            }
        },
    }

    # Test case 1: Missing required 'model' field completely
    slot = Slot(
        name="vehicles_array",
        type="array",
        description="Array of vehicles",
        required=True,
        value=[{"make": "Toyota"}],
        verified=True,
        slot_schema={
            "type": "function",
            "function": {"name": "tool", "parameters": object_schema},
        },
    )

    tool = Tool(dummy_func, "toolname", "desc", [])
    tool.slots = [slot]

    try:
        tool._validate_and_clean_slot_values()
        raise AssertionError("Expected ValidationError to be raised for missing field")
    except ValidationError as e:
        assert "Failed to validate slot values" in str(e)

    # Test case 2: Required field is None (should throw error)
    slot2 = Slot(
        name="vehicles_array",
        type="array",
        description="Array of vehicles",
        required=True,
        value=[{"make": "Toyota", "model": None}],
        verified=True,
        slot_schema={
            "type": "function",
            "function": {"name": "tool", "parameters": object_schema},
        },
    )

    tool2 = Tool(dummy_func, "toolname", "desc", [])
    tool2.slots = [slot2]

    try:
        tool2._validate_and_clean_slot_values()
        raise AssertionError("Expected ValidationError to be raised for None value")
    except ValidationError as e:
        assert "Failed to validate slot values" in str(e)


def test_validate_and_clean_slot_values_empty_string_allowed() -> None:
    """Test that validation allows empty strings for required fields.

    According to the function docstring: 'allow empty values for required fields (e.g. ""), but not None'
    """
    object_schema = {
        "type": "object",
        "required": ["user_info"],
        "properties": {
            "user_info": {
                "type": "object",
                "required": ["name", "email", "phone"],
                "properties": {
                    "name": {"type": "string"},
                    "email": {"type": "string"},
                    "phone": {"type": "string"},
                },
            }
        },
    }

    # Test with empty strings for required fields (should pass)
    slot = Slot(
        name="user_info",
        type="object",
        description="User information",
        required=True,
        value={
            "name": "John Doe",
            "email": "",  # Empty string should be allowed
            "phone": "",  # Empty string should be allowed
        },
        verified=True,
        slot_schema={
            "type": "function",
            "function": {"name": "tool", "parameters": object_schema},
        },
    )

    tool = Tool(dummy_func, "toolname", "desc", [])
    tool.slots = [slot]

    # This should not raise an error - empty strings are allowed
    tool._validate_and_clean_slot_values()

    # Verify the values are preserved
    validated_value = tool.slots[0].value
    assert validated_value["name"] == "John Doe"
    assert validated_value["email"] == ""
    assert validated_value["phone"] == ""


def test_validate_and_clean_slot_values_multiple_slots() -> None:
    """Test validation with multiple slots."""
    object_schema1 = {
        "type": "object",
        "required": ["name"],
        "properties": {"name": {"type": "string"}},
    }

    object_schema2 = {
        "type": "object",
        "required": ["count"],
        "properties": {"count": {"type": "integer"}},
    }

    slot1 = Slot(
        name="name",
        type="string",
        description="Name",
        required=True,
        value="John",
        verified=True,
        slot_schema={
            "type": "function",
            "function": {"name": "tool", "parameters": object_schema1},
        },
    )

    slot2 = Slot(
        name="count",
        type="integer",
        description="Count",
        required=True,
        value=42,
        verified=True,
        slot_schema={
            "type": "function",
            "function": {"name": "tool", "parameters": object_schema2},
        },
    )

    tool = Tool(dummy_func, "toolname", "desc", [])
    tool.slots = [slot1, slot2]
    tool._validate_and_clean_slot_values()

    assert tool.slots[0].value == "John"
    assert tool.slots[1].value == 42


def test_validate_and_clean_slot_values_deeply_nested_objects() -> None:
    """Test validation with deeply nested object structures."""
    object_schema = {
        "type": "object",
        "required": ["data"],
        "properties": {
            "data": {
                "type": "object",
                "required": ["level1"],
                "properties": {
                    "level1": {
                        "type": "object",
                        "required": ["level2"],
                        "properties": {
                            "level2": {
                                "type": "object",
                                "required": ["value"],
                                "properties": {"value": {"type": "string"}},
                            }
                        },
                    }
                },
            }
        },
    }

    slot = Slot(
        name="data",
        type="object",
        description="Nested data",
        required=True,
        value={
            "level1": {"level2": {"value": "deep_value", "extra": "should_be_removed"}}
        },
        verified=True,
        slot_schema={
            "type": "function",
            "function": {"name": "tool", "parameters": object_schema},
        },
    )

    tool = Tool(dummy_func, "toolname", "desc", [])
    tool.slots = [slot]
    tool._validate_and_clean_slot_values()

    validated_value = tool.slots[0].value
    assert validated_value["level1"]["level2"]["value"] == "deep_value"
    assert "extra" not in validated_value["level1"]["level2"]


# Tests for _validate_and_clean_slot_values ------------------------------------------------------------



"""Task graph formatter for Arklex framework.

Formats task definitions into graph structure with nodes, edges, and metadata.
Handles LLM-based intent generation.
"""

from typing import Any

from arklex.utils.logging.logging_utils import LogContext

log_context = LogContext(__name__)


class TaskGraphFormatter:
    """Formats task graphs.

    This class handles formatting of task graphs, ensuring they are properly
    structured and contain all required components.

    Attributes:
        _role (str): Role of the assistant
        _user_objective (str): User's objective
        _builder_objective (str): Builder's objective
        _domain (str): Domain of the task
        _intro (str): Introduction text
        _task_docs (List[Dict[str, Any]]): Task documentation
        _rag_docs (List[Dict[str, Any]]): RAG documentation
        _workers (List[Dict[str, Any]]): List of workers
    """

    # Default worker names - can be overridden in config
    DEFAULT_MESSAGE_WORKER = "MessageWorker"
    DEFAULT_RAG_WORKER = "FaissRAGWorker"
    DEFAULT_SEARCH_WORKER = "SearchWorker"
    DEFAULT_OUTPUT_PROCESS_WORKER = "OutputProcessWorker"

    def __init__(
        self,
        role: str = "",
        user_objective: str = "",
        builder_objective: str = "",
        domain: str = "",
        intro: str = "",
        task_docs: list[dict[str, Any]] | None = None,
        rag_docs: list[dict[str, Any]] | None = None,
        workers: list[dict[str, Any]] | None = None,
        tools: list[dict[str, Any]] | None = None,
        nluapi: str = "",
        slotfillapi: str = "",
        default_intent: str | None = None,
        default_weight: int = 1,
        default_pred: bool = False,
        default_definition: str = "",
        default_sample_utterances: list[str] | None = None,
        nodes: list[Any] | None = None,
        edges: list[Any] | None = None,
        model: object | None = None,
        settings: dict[str, Any] | None = None,
    ) -> None:
        """Initialize the TaskGraphFormatter.

        Args:
            role (str): Role of the assistant
            user_objective (str): User's objective
            builder_objective (str): Builder's objective
            domain (str): Domain of the task
            intro (str): Introduction text
            task_docs (Optional[List[Dict[str, Any]]]): Task documentation
            rag_docs (Optional[List[Dict[str, Any]]]): RAG documentation
            workers (Optional[List[Dict[str, Any]]]): List of workers
            tools (Optional[List[Dict[str, Any]]]): List of tools
            nluapi (str): NLU API
            slotfillapi (str): Slotfill API
            default_intent (Optional[str]): Default intent for edges
            default_weight (int): Default weight for edges
            default_pred (bool): Default pred value for edge attributes
            default_definition (str): Default definition for edge attributes
            default_sample_utterances (Optional[List[str]]): Default sample utterances for edge attributes
            nodes (Optional[List[Any]]): List of nodes
            edges (Optional[List[Any]]): List of edges
            model (Optional[Any]): Language model for intent generation
            settings (Optional[Dict[str, Any]]): Additional configuration settings
        """
        self._role = role
        self._user_objective = user_objective
        self._builder_objective = builder_objective
        self._domain = domain
        self._intro = intro
        self._task_docs = task_docs or []
        self._rag_docs = rag_docs or []
        self._workers = workers or []
        self._tools = tools or []
        self._nluapi = nluapi
        self._slotfillapi = slotfillapi
        self._default_intent = default_intent
        self._default_weight = default_weight
        self._default_pred = default_pred
        self._default_definition = default_definition
        self._default_sample_utterances = default_sample_utterances or []
        self._nodes = nodes
        self._edges = edges
        self._model = model
        self._settings = settings

    def _find_worker_by_name(self, worker_name: str) -> dict[str, str]:
        """Get worker info from name, with fallback mappings.

        Args:
            worker_name (str): Name of the worker to find info for

        Returns:
            Dict[str, str]: Worker info with 'id' and 'name' keys
        """
        if self._workers:
            for worker in self._workers:
                if isinstance(worker, dict) and worker.get("name") == worker_name:
                    return {
                        "id": worker.get("id", worker_name.lower()),
                        "name": worker.get("name", worker_name),
                    }

        # Fallback mappings based on config
        fallback_workers = {
            "MessageWorker": {
                "id": "26bb6634-3bee-417d-ad75-23269ac17bc3",
                "name": "MessageWorker",
            },
            "FaissRAGWorker": {"id": "FaissRAGWorker", "name": "FaissRAGWorker"},
            "SearchWorker": {
                "id": "9c15af81-04b3-443e-be04-a3522124b905",
                "name": "SearchWorker",
            },
            "OutputProcessWorker": {
                "id": "OutputProcessWorker",
                "name": "OutputProcessWorker",
            },
        }
        return fallback_workers.get(
            worker_name, {"id": worker_name.lower(), "name": worker_name}
        )

    def format_task_graph(self, tasks: list[dict[str, Any]]) -> dict[str, Any]:
        """Format tasks into complete graph structure with nodes, edges, and metadata.

        Args:
            tasks (List[Dict[str, Any]]): List of task definitions to format

        Returns:
            Dict[str, Any]: Complete task graph with nodes, edges, and metadata
        """
        if self._nodes is not None and self._edges is not None:
            return {"nodes": self._nodes, "edges": self._edges}

        # Format nodes and edges
        nodes, node_lookup, all_task_node_ids = self._format_nodes(tasks)
        start_node_id = "0"  # Start node is always "0" in our current implementation
        edges = self._format_edges(tasks, node_lookup, all_task_node_ids, start_node_id)

        # Process node values
        for _, node_data in nodes:
            # If the value is a dict, replace with its description
            value = node_data.get("attribute", {}).get("value")
            if isinstance(value, dict):
                desc = value.get("description")
                if desc:
                    node_data["attribute"]["value"] = desc
                else:
                    node_data["attribute"]["value"] = str(value)
            # If the value is a list or nested, flatten to string
            elif isinstance(value, list):
                node_data["attribute"]["value"] = ", ".join(str(v) for v in value)

        reusable_tasks = {}  # Define reusable_tasks, can be passed in later

        graph = {
            "nodes": nodes,
            "edges": edges,
            "tasks": tasks,  # Include tasks for connectivity logic
            "role": self._role,
            "user_objective": self._user_objective,
            "builder_objective": self._builder_objective,
            "domain": self._domain,
            "intro": self._intro,
            "task_docs": self._task_docs,
            "rag_docs": self._rag_docs,
            "workers": self._workers,
            "tools": self._tools,
            "nluapi": self._nluapi,
            "slotfillapi": self._slotfillapi,
            "reusable_tasks": reusable_tasks,
            "settings": self._settings,
        }

        return graph

    def _format_nodes(self, tasks: list[dict]) -> tuple[list, dict, list]:
        """Create nodes for start, tasks, and steps with worker assignments.

        Args:
            tasks (List[Dict]): List of task definitions to create nodes for

        Returns:
            Tuple[List, Dict, List]: (nodes, node_lookup, all_task_node_ids)
                - nodes: List of formatted node data
                - node_lookup: Mapping of task identifiers to node IDs
                - all_task_node_ids: List of all task node IDs
        """
        nodes = []
        node_id_counter = 0
        node_lookup = {}
        all_task_node_ids = []

        # Create start node
        start_node_id = str(node_id_counter)
        message_worker_id = self._find_worker_by_name(self.DEFAULT_MESSAGE_WORKER)
        nodes.append(
            [
                start_node_id,
                {
                    "resource": {
                        "id": message_worker_id["id"],
                        "name": message_worker_id["name"],
                    },
                    "attribute": {
                        "value": "Hello! I'm here to assist you with any customer service inquiries.",
                        "task": "start message",
                        "directed": False,
                    },
                    "limit": 1,
                    "type": "start",
                },
            ]
        )
        node_id_counter += 1

        # First pass: Create all task and step nodes
        for task_idx, task in enumerate(tasks):
            # Use 'id' if available, otherwise generate one
            task_identifier = task.get("id", f"task_{task_idx}")

            # Ensure task has an id field
            if "id" not in task:
                task["id"] = task_identifier

            # Use the resource from the task if available, otherwise default to MessageWorker
            resource_name = self.DEFAULT_MESSAGE_WORKER
            if task.get("resource") and isinstance(task["resource"], dict):
                resource_name = task["resource"].get(
                    "name", self.DEFAULT_MESSAGE_WORKER
                )
            elif task.get("resource"):
                resource_name = str(task["resource"])

            resource_info = self._find_worker_by_name(resource_name)

            task_node_id = str(node_id_counter)
            node_data = {
                "resource": {"id": resource_info["id"], "name": resource_info["name"]},
                "attribute": {
                    "value": task.get("description", ""),
                    "task": task.get("name", ""),
                    "directed": False,
                },
            }
            if task.get("type"):
                node_data["type"] = task.get("type")
            if task.get("limit"):
                node_data["limit"] = task.get("limit")

            nodes.append([task_node_id, node_data])
            node_lookup[task_identifier] = task_node_id
            all_task_node_ids.append(task_node_id)
            node_id_counter += 1

            # Ensure steps are properly broken down
            steps = task.get("steps", [])
            if not steps and task.get("description"):
                # If no steps defined, create a single step from the task description
                steps = [
                    {"description": task.get("description", ""), "step_id": "step_1"}
                ]
                task["steps"] = steps

            for step_idx, step in enumerate(steps):
                step_id = f"{task_identifier}_step{step_idx}"
                step_node_id = str(node_id_counter)

                # Use the resource from the step if available (from best practice manager)
                step_worker_name = self.DEFAULT_MESSAGE_WORKER
                if (
                    isinstance(step, dict)
                    and step.get("resource")
                    and isinstance(step["resource"], dict)
                ):
                    step_worker_name = step["resource"].get(
                        "name", self.DEFAULT_MESSAGE_WORKER
                    )
                elif isinstance(step, dict) and step.get("resource"):
                    step_worker_name = str(step["resource"])

                step_worker_info = self._find_worker_by_name(step_worker_name)

                # Simplify step value to use simple string instead of complex nested structure
                if isinstance(step, dict):
                    # Check if step has a complex structure with task, description, step_id, etc.
                    if "task" in step and "description" in step and "step_id" in step:
                        # This is a complex step structure, extract just the description
                        step_value = step.get("description", "")
                    else:
                        # Simple step structure, use description or task
                        step_value = step.get("description", step.get("task", ""))
                elif isinstance(step, str):
                    step_value = step
                else:
                    step_value = str(step)

                nodes.append(
                    [
                        step_node_id,
                        {
                            "resource": {
                                "id": step_worker_info["id"],
                                "name": step_worker_info["name"],
                            },
                            "attribute": {
                                "value": step_value,
                                "task": task.get("name", ""),
                                "directed": False,
                            },
                        },
                    ]
                )
                node_lookup[step_id] = step_node_id
                node_id_counter += 1

        return nodes, node_lookup, all_task_node_ids

    def _create_edge_attributes(
        self,
        intent: str | None = None,
        weight: int = 1,
        pred: bool = False,
        definition: str = "",
        sample_utterances: list[str] | None = None,
    ) -> dict[str, Any]:
        """Create edge attributes with intent and metadata.

        Args:
            intent (Optional[str]): Intent for the edge, defaults to None
            weight (int): Edge weight, defaults to 1
            pred (bool): Prediction flag, defaults to False
            definition (str): Edge definition, defaults to empty string
            sample_utterances (Optional[List[str]]): Sample utterances for the edge

        Returns:
            Dict[str, Any]: Edge attributes dictionary with intent and metadata
        """
        return {
            "intent": intent,
            "attribute": {
                "weight": weight,
                "pred": pred,
                "definition": definition,
                "sample_utterances": sample_utterances or [],
            },
        }

    def _format_edges(
        self,
        tasks: list[dict[str, Any]],
        node_lookup: dict[str, str],
        all_task_node_ids: list[str],
        start_node_id: str,
    ) -> list[Any]:
        """Create edges between nodes with LLM-generated intents.

        Args:
            tasks (List[Dict[str, Any]]): List of task definitions
            node_lookup (Dict[str, str]): Mapping of task identifiers to node IDs
            all_task_node_ids (List[str]): List of all task node IDs
            start_node_id (str): ID of the start node

        Returns:
            List[Any]: List of formatted edge data
        """
        edges = []

        # Generate descriptive intents for main task edges using LLM
        for task in tasks:
            task_identifier = task.get("id", f"task_{tasks.index(task)}")
            task_node_id = node_lookup.get(task_identifier)

            if task_node_id:
                dependencies = task.get("dependencies", [])
                if not dependencies:
                    # This is a main task edge from the start node.
                    # Use the intent from the task, or generate a descriptive one
                    intent = task.get("intent", "")
                    if not intent:
                        # Generate a descriptive intent based on task name and description
                        task_name = task.get("name", "")
                        task_description = task.get("description", "")

                        # Create a prompt for intent generation
                        intent_prompt = f"""
                        Given a task with the following details, generate a user-facing intent that describes what the user wants to accomplish.
                        
                        Task Name: {task_name}
                        Task Description: {task_description}
                        
                        Generate a natural, user-facing intent that describes what the user is trying to achieve. 
                        Examples of good intents:
                        - "User inquires about purchasing options"
                        - "User wants to explore rental options"
                        - "User asks about delivery times and logistics"
                        - "User has technical support or troubleshooting queries"
                        - "User seeks information on robot capabilities and features"
                        - "User is interested in customization options"
                        - "User inquires about service and maintenance"
                        - "User wants to learn about new releases and updates"
                        
                        Intent: """

                        try:
                            # Use the model to generate intent if available
                            if self._model:
                                from langchain_core.messages import HumanMessage

                                response = self._model.invoke(
                                    [HumanMessage(content=intent_prompt)]
                                )
                                intent = response.content.strip().strip('"').strip("'")
                                # Clean up the response to ensure it's a valid intent
                                if not intent or intent.lower() in [
                                    "none",
                                    "null",
                                    "n/a",
                                ]:
                                    raise ValueError("Invalid intent generated")
                            else:
                                intent = "User inquires about purchasing options"  # default fallback
                        except Exception as e:
                            log_context.warning(
                                f"Failed to generate intent for task {task_name}: {e}"
                            )
                            intent = (
                                "User inquires about purchasing options"  # fallback
                            )

                    edges.append(
                        [
                            start_node_id,
                            task_node_id,
                            self._create_edge_attributes(intent=intent, pred=True),
                        ]
                    )
                else:
                    # This handles dependencies between tasks.
                    for dep in dependencies:
                        # Handle edge cases where dependency is None or not a string/dict
                        if dep is None:
                            log_context.warning("Skipping None dependency")
                            continue
                        elif isinstance(dep, str):
                            dep_id = dep
                        elif isinstance(dep, dict):
                            dep_id = dep.get("id")
                            if dep_id is None:
                                log_context.warning(
                                    "Skipping dependency dict without 'id' field"
                                )
                                continue
                        else:
                            log_context.warning(
                                f"Skipping invalid dependency type: {type(dep)}"
                            )
                            continue

                        # Find the source task from the list of tasks
                        source_task = next(
                            (t for t in tasks if t.get("id") == dep_id), None
                        )

                        if source_task:
                            source_steps = source_task.get("steps", [])
                            if source_steps:
                                # Dependency should come from the last step of the source task
                                last_step_identifier = (
                                    f"{dep_id}_step{len(source_steps) - 1}"
                                )
                                source_node_id = node_lookup.get(last_step_identifier)
                            else:
                                # If no steps, dependency is from the task node itself
                                source_node_id = node_lookup.get(dep_id)
                        else:
                            # Fallback to original logic if source task not found
                            source_node_id = node_lookup.get(dep_id)

                        if source_node_id:
                            edges.append(
                                [
                                    source_node_id,
                                    task_node_id,
                                    self._create_edge_attributes(),
                                ]
                            )
                        else:
                            log_context.warning(
                                f"Could not find source node for dependency '{dep_id}'"
                            )

        # Steps
        for task in tasks:
            task_identifier = task.get("id", f"task_{tasks.index(task)}")
            task_node_id = node_lookup.get(task_identifier)

            if task_node_id:
                steps = task.get("steps", [])
                if steps:
                    first_step_id = f"{task_identifier}_step0"
                    if first_step_id in node_lookup:
                        edges.append(
                            [
                                task_node_id,
                                node_lookup[first_step_id],
                                self._create_edge_attributes(intent=None),
                            ]
                        )
                        for i in range(len(steps) - 1):
                            current_step_id = f"{task_identifier}_step{i}"
                            next_step_id = f"{task_identifier}_step{i + 1}"
                            if (
                                current_step_id in node_lookup
                                and next_step_id in node_lookup
                            ):
                                edges.append(
                                    [
                                        node_lookup[current_step_id],
                                        node_lookup[next_step_id],
                                        self._create_edge_attributes(intent=None),
                                    ]
                                )

        return edges



from typing import Any
from unittest.mock import MagicMock, patch

from langchain_core.messages import AIMessage

from arklex.resources.resource_types import WorkerItem
from integration_tests.utils.base import BaseTestOrchestrator, ChatRole


def mock_model_invoke(messages: list[Any]) -> AIMessage:
    # Get the last message content
    last_message = messages[-1]
    content = last_message.content

    # RAG question reformulation requests
    if "formulate a standalone question" in content:
        if "What products do you offer?" in content:
            return AIMessage(content="What products do you offer?")
        elif "What is your company's culture?" in content:
            return AIMessage(content="What is your company's culture?")
        else:
            raise ValueError(
                f"Unknown RAG question reformulation request: {content[-100:]}"
            )

    # Retriever needed requests
    elif "decide whether information retrieval is needed" in content:
        if "What products do you offer?" in content:  # noqa: SIM114
            return AIMessage(content="yes")
        elif "What is your company's culture?" in content:  # noqa: SIM114
            return AIMessage(content="yes")
        else:
            raise ValueError(f"Unknown retriever needed request: {content[-100:]}")

    return AIMessage(content="This is a demo worker response.")


def mock_structured_output_invoke(messages: list[Any]) -> dict[str, str]:
    # Get the last message content
    last_message = messages[-1]
    content = last_message.content

    # Intent detection responses
    if "Given the following intents and their definitions" in content:
        if content.endswith("How is the weather?\n"):
            return {"intent": "ask about weather"}
        elif content.endswith("Which car would you like to buy?\n"):
            return {"intent": "ask about car options"}
        elif content.endswith("What products do you offer?\n"):
            return {"intent": "ask about company products"}
        elif content.endswith("What is your company's culture?\n"):
            return {"intent": "ask about company culture"}
        elif content.endswith("Connect me with a human agent\n"):
            return {"intent": "user want to connect with a human agent"}
        else:
            raise ValueError(f"Unknown intent detection request: {content[-100:]}")

    # Default response for other structured outputs
    return {"intent": "others"}


def create_mock_llm() -> MagicMock:
    mock_llm = MagicMock()

    mock_llm.invoke = MagicMock(side_effect=mock_model_invoke)

    mock_structured_llm = MagicMock()
    mock_structured_llm.invoke = MagicMock(side_effect=mock_structured_output_invoke)
    mock_llm.with_structured_output = MagicMock(return_value=mock_structured_llm)

    return mock_llm


@patch("arklex.models.model_service.load_llm")
async def test_workers(mock_load_llm: MagicMock) -> None:
    # Configure load_llm to return our mock LLM
    mock_load_llm.return_value = create_mock_llm()

    orchestrator = BaseTestOrchestrator(
        "integration_tests/taskgraphs/worker_taskgraph.json"
    )
    params = BaseTestOrchestrator.init_params()
    chat_history, params = params["chat_history"], params["parameters"]

    # start message (direct message)
    text = "<start>"
    output = await orchestrator.get_response(text, chat_history, params)
    chat_history.append({"role": ChatRole.USER, "content": text})
    chat_history.append({"role": ChatRole.ASSISTANT, "content": output["answer"]})
    params = output["parameters"]
    assert (
        output["answer"]
        == "Hello! I'm here to assist you with any customer service inquiries."
    )

    # message worker (undirected message)
    text = "How is the weather?"
    output = await orchestrator.get_response(text, chat_history, params)
    chat_history.append({"role": ChatRole.USER, "content": text})
    chat_history.append({"role": ChatRole.ASSISTANT, "content": output["answer"]})
    params = output["parameters"]
    assert len(output["answer"]) > 1

    # multiple choice worker
    text = "Which car would you like to buy?"
    output = await orchestrator.get_response(text, chat_history, params)
    chat_history.append({"role": ChatRole.USER, "content": text})
    chat_history.append({"role": ChatRole.ASSISTANT, "content": output["answer"]})
    params = output["parameters"]
    assert output["answer"] == "Which car would you like to buy?"
    assert output["choice_list"] == ["Car A", "Car B", "Car C"]

    # Milvus RAG worker
    text = "What products do you offer?"
    output = await orchestrator.get_response(text, chat_history, params)
    chat_history.append({"role": ChatRole.USER, "content": text})
    chat_history.append({"role": ChatRole.ASSISTANT, "content": output["answer"]})
    params = output["parameters"]
    last_trajectory = params["memory"]["trajectory"][-1][0]
    assert last_trajectory["info"]["resource"]["id"] == WorkerItem.MILVUS_RAG_WORKER
    assert len(last_trajectory["steps"]) > 0

    # RAG message worker
    text = "What is your company's culture?"
    output = await orchestrator.get_response(text, chat_history, params)
    chat_history.append({"role": ChatRole.USER, "content": text})
    chat_history.append({"role": ChatRole.ASSISTANT, "content": output["answer"]})
    params = output["parameters"]
    last_trajectory = params["memory"]["trajectory"][-1][0]
    assert last_trajectory["info"]["resource"]["id"] == WorkerItem.RAG_MESSAGE_WORKER
    assert len(last_trajectory["steps"]) > 0

    # Human in the loop worker
    text = "Connect me with a human agent"
    output = await orchestrator.get_response(text, chat_history, params)
    chat_history.append({"role": ChatRole.USER, "content": text})
    chat_history.append({"role": ChatRole.ASSISTANT, "content": output["answer"]})
    params = output["parameters"]
    assert output["answer"] == "I'll connect you to a representative!"
    assert output["human_in_the_loop"] == "live"



"""Search worker implementation for the Arklex framework.

This module provides a specialized worker for handling search-related tasks. It implements
a search engine worker that can answer user questions based on real-time online search
results. The worker uses a state graph to manage the workflow of search operations and
response generation, integrating with the framework's tool generation system.
"""

from typing import Any

from arklex.orchestrator.entities.orchestrator_state_entities import (
    OrchestratorState,
    StatusEnum,
)
from arklex.resources.tools.rag.search import SearchEngine
from arklex.resources.tools.utils import ToolGenerator
from arklex.resources.workers.base.base_worker import BaseWorker
from arklex.resources.workers.search.entities import (
    SearchWorkerData,
    SearchWorkerOutput,
)
from arklex.utils.logging.logging_utils import LogContext

log_context = LogContext(__name__)


class SearchWorker(BaseWorker):
    description: str = (
        "Answer the user's questions based on real-time online search results"
    )

    def __init__(self) -> None:
        super().__init__()

    def init_worker_data(
        self, orch_state: OrchestratorState, node_specific_data: dict[str, Any]
    ) -> None:
        self.orch_state = orch_state
        self.search_worker_data = SearchWorkerData(**node_specific_data)

    def _execute(self) -> SearchWorkerOutput:
        search_engine: SearchEngine = SearchEngine()
        retrieved_text = search_engine.search(
            chat_history=self.orch_state.user_message.history,
            bot_config=self.orch_state.bot_config,
        )
        self.orch_state.message_flow = retrieved_text
        response = ToolGenerator.context_generate(self.orch_state)
        return SearchWorkerOutput(
            response=response,
            status=StatusEnum.COMPLETE,
        )



from typing import Any

from pydantic import BaseModel

from arklex.orchestrator.entities.orchestrator_state_entities import (
    StatusEnum,
)
from arklex.resources.workers.base.entities import WorkerOutput


class MilvusRAGWorkerData(BaseModel):
    """Data for the Milvus RAG worker."""

    bot_id: str
    version: str
    collection_name: str
    tags: dict[str, Any]
    possible_tags: dict[str, list[str]] | None = None


class MilvusRAGWorkerOutput(WorkerOutput):
    """Response for the Milvus RAG worker."""

    response: str
    status: StatusEnum



import re

from langchain_core.prompts import PromptTemplate

from arklex.memory.entities.memory_entities import ResourceRecord
from arklex.models.model_service import ModelService
from arklex.orchestrator.entities.orchestrator_state_entities import OrchestratorState
from arklex.orchestrator.executor.entities import NodeResponse
from arklex.resources.resource_types import WorkerItem
from arklex.utils.logging.logging_utils import LogContext
from arklex.utils.prompts import load_prompts

log_context = LogContext(__name__)

RAG_NODES_STEPS = {
    WorkerItem.MESSAGE_WORKER.value: "message",
    WorkerItem.FAISS_RAG_WORKER.value: "faiss_retrieve",
    WorkerItem.MILVUS_RAG_WORKER.value: "milvus_retrieve",
    WorkerItem.RAG_MESSAGE_WORKER.value: "milvus_retrieve",
}


def post_process_response(
    orch_state: OrchestratorState,
    node_response: NodeResponse,
) -> NodeResponse:
    """
    Post-processes the chatbot's response to ensure content quality and determine whether human takeover is needed.

    This function performs the following steps:
    1. **Link Validation**: Compares links in the bot's response against links present in the context.
    If the response includes invalid links, they are removed and the response is optionally regenerated via LLM.
    2. **HITL Proposal Trigger**: If HITL proposal is enabled and a HITL worker is available, determines
    whether to suggest a handoff to a human assistant based on confidence and relevance heuristics.

    Args:
        orch_state (OrchestratorState): Current state of the conversation including response, context, and metadata.
        node_response (NodeResponse): Response from the current node.

    Returns:
        NodeResponse: The updated node response with potentially cleaned or rephrased response,
                    and possibly a human handoff suggestion.
    """
    context_links = _build_context(orch_state.sys_instruct, orch_state.trajectory)
    response_links = _extract_links(node_response.response)
    missing_links = response_links - context_links
    if missing_links:
        log_context.info(
            f"Some answer links are NOT present in the context. Missing: {missing_links}"
        )
        node_response.response = _remove_invalid_links(
            node_response.response, missing_links
        )
        node_response.response = _rephrase_answer(orch_state, node_response.response)

    return node_response


def _build_context(sys_instruct: str, trajectory: list[list[ResourceRecord]]) -> set:
    context_links = _extract_links(sys_instruct)
    for resource_group in trajectory:
        for resource in resource_group:
            if _include_resource(resource):
                context_links.update(_extract_links(resource.output))
            rag_step_type = RAG_NODES_STEPS.get(
                resource.info.get("resource", {}).get("id")
            )
            if rag_step_type:
                for step in resource.steps:
                    try:
                        if rag_step_type in step:
                            step_links = _extract_links_from_nested_dict(
                                step[rag_step_type]
                            )
                            context_links.update(step_links)
                    except Exception as e:
                        log_context.warning(
                            f"Error extracting links from step: {e} — step: {step}"
                        )
    return context_links


def _include_resource(resource: ResourceRecord) -> bool:
    """Determines whether a ResourceRecord's output should be included in context.

    Excludes any output where a 'context_generate' flag is present in steps.
    """
    return not any(step.get("context_generate") for step in resource.steps)


def _extract_links(text: str) -> set:
    markdown_links = re.findall(r"\[[^\]]+\]\((https?://[^\s)]+)\)", text)
    cleaned_text = re.sub(r"\[[^\]]+\]\((https?://[^\s)]+)\)", "", text)
    raw_links = re.findall(r"(?:https?://|www\.)[^\s)\"']+", cleaned_text)
    all_links = set(markdown_links + raw_links)
    return {link.rstrip(".,;)!?\"'") for link in all_links}


def _extract_links_from_nested_dict(step: dict | list | str) -> set:
    links = set()

    def _recurse(val: dict | list | str) -> None:
        if isinstance(val, str):
            links.update(_extract_links(val))
        elif isinstance(val, dict):
            for v in val.values():
                _recurse(v)
        elif isinstance(val, list):
            for item in val:
                _recurse(item)

    _recurse(step)
    return links


def _remove_invalid_links(response: str, links: set) -> str:
    sorted_links = sorted([re.escape(link) for link in links], key=len, reverse=True)
    links_regex = "|".join(sorted_links)
    cleaned_response = re.sub(links_regex, "", response)
    return re.sub(r"\s+", " ", cleaned_response).strip()


def _rephrase_answer(orch_state: OrchestratorState, response: str) -> str:
    """Rephrases the answer using an LLM after link removal."""
    model_service = ModelService(orch_state.bot_config.llm_config)
    prompt: PromptTemplate = PromptTemplate.from_template(
        load_prompts(orch_state.bot_config.language)["regenerate_response"]
    )
    input_prompt = prompt.invoke(
        {
            "sys_instruct": orch_state.sys_instruct,
            "original_answer": response,
            "formatted_chat": orch_state.user_message.history,
        }
    )
    log_context.info(f"Prompt: {input_prompt.text}")
    answer: str = model_service.get_response(input_prompt.text)
    return answer



from typing import Any

from agents import (
    Agent,
    HandoffOutputItem,
    InputGuardrailTripwireTriggered,
    ItemHelpers,
    MessageOutputItem,
    OutputGuardrailTripwireTriggered,
    Runner,
    ToolCallItem,
    ToolCallOutputItem,
)
from openai.types.responses import ResponseTextDeltaEvent
from pydantic import BaseModel, Field

from arklex.orchestrator.entities.orchestrator_state_entities import (
    OrchestratorState,
)
from arklex.orchestrator.types.stream_types import EventType, StreamType
from arklex.resources.agents.base.agent import BaseAgent, register_agent
from arklex.resources.agents.base.entities import PromptVariable
from arklex.resources.agents.llm_based_agent.guardrail_agent import (
    create_input_guardrail_function,
    create_output_guardrail_function,
)
from arklex.utils.logging.logging_utils import LogContext

log_context = LogContext(__name__)


class OpenAIAgentData(BaseModel):
    """Data for the OpenAIAgent."""

    name: str
    prompt: str
    prompt_variables: list[PromptVariable] = []
    start_agent: bool = False
    agent_start_message: str | None = None
    handoff_description: str | None = None
    safety_response: str | None = None


class OpenAIAgentOutput(BaseModel):
    """Output for the OpenAIAgent."""

    response: str
    last_agent_name: str
    tool_calls: list[dict[str, Any]]


class ExecuteParams(BaseModel):
    start_message: str | None
    safety_response: str | None
    input_guardrails: list = Field(default_factory=list)
    output_guardrails: list = Field(default_factory=list)


SAFETY_RESPONSE = "An error occurred while processing your request. Please try again."


@register_agent
class OpenAIAgent(BaseAgent):
    description: str = "General-purpose Arklex agent for chat or voice."

    def __init__(
        self, agent: Agent, state: OrchestratorState, **kwargs: object
    ) -> None:
        self.agent = agent
        self.state = state
        self.last_agent_name = ""
        self.execute_params = ExecuteParams.model_validate(kwargs)
        self.configure_agent(
            self.execute_params.input_guardrails, self.execute_params.output_guardrails
        )

    def configure_agent(
        self, input_guardrails: list[Agent], output_guardrails: list[Agent]
    ) -> None:
        self.agent.input_guardrails = [
            create_input_guardrail_function(input_guardrail.name, input_guardrail)
            for input_guardrail in input_guardrails
        ]
        log_context.info(
            f"Input guardrails: {[input_guardrail.name for input_guardrail in input_guardrails]} configured for agent {self.agent.name}"
        )
        self.agent.output_guardrails = [
            create_output_guardrail_function(output_guardrail.name, output_guardrail)
            for output_guardrail in output_guardrails
        ]
        log_context.info(
            f"Output guardrails: {[output_guardrail.name for output_guardrail in output_guardrails]} configured for agent {self.agent.name}"
        )

    async def response(
        self, trajectory: list[dict[str, Any]]
    ) -> tuple[str, list[dict[str, Any]]]:
        final_response = ""
        result: list = []
        is_error = False
        tool_calls = []
        try:
            call_name_map = {}
            result = await Runner.run(self.agent, trajectory)
            for new_item in result.new_items:
                agent_name = new_item.agent.name
                if isinstance(new_item, MessageOutputItem):
                    final_response = ItemHelpers.text_message_output(new_item)
                    log_context.info(f"{agent_name}: {final_response}")
                elif isinstance(new_item, HandoffOutputItem):
                    log_context.info(
                        f"Handed off from {agent_name} to {new_item.target_agent.name}"
                    )
                elif isinstance(new_item, ToolCallItem):
                    call_id = new_item.raw_item.call_id
                    tool_name = new_item.raw_item.name
                    log_context.info(
                        f"{agent_name}: Calling tool '{tool_name}' (id={call_id})"
                    )
                    # Store mapping for later use (when output arrives)
                    call_name_map[call_id] = tool_name
                    tool_call_msg = {
                        "type": new_item.type,
                        "raw_item": new_item.raw_item.model_dump(),
                    }
                    tool_calls.append(tool_call_msg)
                elif isinstance(new_item, ToolCallOutputItem):
                    log_context.info(
                        f"{agent_name} tool call output: {new_item.output}"
                    )
                    call_id = new_item.raw_item.get("call_id")
                    tool_name = call_name_map.get(call_id, "unknown_tool")
                    tool_call_output_msg = {
                        "type": new_item.type,
                        "name": tool_name,
                        "response": new_item.output,
                        "raw_item": dict(new_item.raw_item),
                    }
                    tool_calls.append(tool_call_output_msg)
                else:
                    log_context.info(f"{agent_name} unknown item: {new_item}")
        except InputGuardrailTripwireTriggered as e:
            log_context.error(f"Input guardrail tripped: {e}")
            is_error = True
        except OutputGuardrailTripwireTriggered as e:
            log_context.error(f"Output guardrail tripped: {e}")
            is_error = True
        except Exception as e:
            log_context.error(f"Error during agent execution: {e}")
            is_error = True

        if not is_error:
            new_traj = result.to_input_list()
        else:
            agent_name = self.agent.name or ""
            new_traj = trajectory.copy()
            safety_response = self.execute_params.safety_response or SAFETY_RESPONSE
            new_traj.append({"role": "assistant", "content": safety_response})
            final_response = safety_response

        self.last_agent_name = agent_name
        return final_response, new_traj, tool_calls

    async def stream_response(
        self, trajectory: list[dict[str, Any]]
    ) -> tuple[str, list[dict[str, Any]]]:
        final_response = ""
        is_error = False
        result = Runner.run_streamed(self.agent, trajectory)
        tool_calls = []
        try:
            call_name_map = {}
            async for event in result.stream_events():
                # raw final response for streaming
                if event.type == "raw_response_event" and isinstance(
                    event.data, ResponseTextDeltaEvent
                ):
                    await self.state.message_queue.put(
                        {
                            "event": EventType.CHUNK.value,
                            "message_chunk": event.data.delta,
                        }
                    )
                elif event.type == "run_item_stream_event":
                    new_item = event.item
                    agent_name = new_item.agent.name
                    if isinstance(new_item, MessageOutputItem):
                        final_response = ItemHelpers.text_message_output(new_item)
                        log_context.info(f"{agent_name}: {final_response}")
                    elif isinstance(new_item, HandoffOutputItem):
                        log_context.info(
                            f"Handed off from {new_item.source_agent.name} to {new_item.target_agent.name}"
                        )
                    elif isinstance(new_item, ToolCallItem):
                        call_id = new_item.raw_item.call_id
                        tool_name = new_item.raw_item.name
                        log_context.info(
                            f"{agent_name}: Calling tool '{tool_name}' (id={call_id})"
                        )
                        # Store mapping for later use (when output arrives)
                        call_name_map[call_id] = tool_name
                        tool_call_msg = {
                            "type": new_item.type,
                            "raw_item": new_item.raw_item.model_dump(),
                        }
                        tool_calls.append(tool_call_msg)
                        await self.state.message_queue.put(
                            {"event": EventType.TOOL_CALL.value, **tool_call_msg}
                        )
                    elif isinstance(new_item, ToolCallOutputItem):
                        log_context.info(
                            f"{agent_name} tool call output: {new_item.output}"
                        )
                        call_id = new_item.raw_item.get("call_id")
                        tool_name = call_name_map.get(call_id, "unknown_tool")
                        tool_call_output_msg = {
                            "type": new_item.type,
                            "name": tool_name,
                            "response": new_item.output,
                            "raw_item": dict(new_item.raw_item),
                        }
                        tool_calls.append(tool_call_output_msg)
                        await self.state.message_queue.put(
                            {
                                "event": EventType.TOOL_CALL_OUTPUT.value,
                                **tool_call_output_msg,
                            }
                        )
                    else:
                        log_context.info(f"{agent_name} unknown item: {new_item}")
        except InputGuardrailTripwireTriggered as e:
            log_context.error(f"Input guardrail tripped: {e}")
            is_error = True
        except OutputGuardrailTripwireTriggered as e:
            log_context.error(f"Output guardrail tripped: {e}")
            is_error = True
        except Exception as e:
            log_context.error(f"Error during agent execution: {e}")
            is_error = True

        if not is_error:
            new_traj = result.to_input_list()
        else:
            agent_name = self.agent.name or ""
            new_traj = trajectory.copy()
            safety_response = self.execute_params.safety_response or SAFETY_RESPONSE
            await self.state.message_queue.put(
                {
                    "event": EventType.CHUNK.value,
                    "message_chunk": safety_response,
                }
            )
            new_traj.append({"role": "assistant", "content": safety_response})
            final_response = safety_response

        self.last_agent_name = agent_name
        return final_response, new_traj, tool_calls

    async def execute(self) -> tuple[OrchestratorState, OpenAIAgentOutput]:
        user_message = self.state.user_message.message
        trajectory = self.state.openai_agents_trajectory.copy() or []

        if user_message == "<start>":
            if (
                self.execute_params.start_message
                and self.execute_params.start_message.strip()
            ):
                trajectory.append(
                    {"role": "assistant", "content": self.execute_params.start_message}
                )
                self.state.openai_agents_trajectory = trajectory
                return self.state, OpenAIAgentOutput(
                    response=self.execute_params.start_message,
                    last_agent_name="",
                    tool_calls=[],
                )
            log_context.info("No start message configured for agent")
        trajectory.append({"role": "user", "content": user_message})

        if self.state.stream_type == StreamType.NON_STREAM:
            response, new_traj, tool_calls = await self.response(trajectory)
        else:
            response, new_traj, tool_calls = await self.stream_response(trajectory)
        self.state.openai_agents_trajectory = new_traj
        return self.state, OpenAIAgentOutput(
            response=response,
            last_agent_name=self.last_agent_name,
            tool_calls=tool_calls,
        )



from pydantic import BaseModel

from arklex.orchestrator.entities.orchestrator_state_entities import StatusEnum
from arklex.resources.workers.base.entities import WorkerOutput


class SearchWorkerData(BaseModel):
    """Data for the search worker."""


class SearchWorkerOutput(WorkerOutput):
    """Response for the search worker."""

    response: str
    status: StatusEnum



import copy
import time
from typing import Any

import janus
from langchain_core.runnables import RunnableLambda
from pydantic import BaseModel

from arklex.memory.entities.memory_entities import ResourceRecord
from arklex.models.llm_config import LLMConfig
from arklex.models.model_service import ModelService
from arklex.orchestrator.entities.orchestrator_param_entities import OrchestratorParams
from arklex.orchestrator.entities.orchestrator_state_entities import (
    BotConfig,
    ConvoMessage,
    OrchestratorState,
    StatusEnum,
)
from arklex.orchestrator.entities.taskgraph_entities import NodeInfo
from arklex.orchestrator.executor.entities import NodeResponse
from arklex.orchestrator.executor.executor import Executor
from arklex.orchestrator.task_graph.nlu_graph import NLUGraph
from arklex.orchestrator.types.stream_types import StreamType
from arklex.resources.agents.base.agent import BaseAgent, register_agent
from arklex.resources.agents.rule_based_agent.post_process import post_process_response
from arklex.resources.tools.utils import ToolGenerator
from arklex.utils.logging.logging_utils import LogContext
from arklex.utils.prompts import load_prompts
from arklex.utils.utils import format_chat_history

log_context = LogContext(__name__)


class NLUAgentData(BaseModel):
    prompt: str
    response_length: int
    language: str


@register_agent
class NLUAgent(BaseAgent):
    def __init__(
        self,
        llm_config: LLMConfig,
        nlu_graph: NLUGraph,
        executor: Executor,
    ) -> None:
        self.user_prefix = "user"
        self.llm_config = llm_config
        self.executor = executor
        self.nlu_graph = nlu_graph
        self.agent_data: NLUAgentData = NLUAgentData.model_validate(
            self.nlu_graph.agent_node.data
        )

    def format_system_prompt(self) -> str:
        if self.agent_data.language == "EN":
            return (
                self.agent_data.prompt
                + f"\nLimit the response within {self.agent_data.response_length} words"
            )
        else:
            return (
                self.agent_data.prompt
                + f"限制回复长度在{self.agent_data.response_length}字以内"
            )

    def init_params(
        self, inputs: dict[str, Any]
    ) -> tuple[str, str, OrchestratorParams, OrchestratorState]:
        text: str = inputs["text"]
        chat_history: list[dict[str, str]] = inputs["chat_history"]
        input_params: dict[str, Any] | None = inputs["parameters"]

        params: OrchestratorParams = OrchestratorParams()

        if input_params:
            params = OrchestratorParams.model_validate(input_params)

        chat_history_copy: list[dict[str, str]] = copy.deepcopy(chat_history)
        chat_history_copy.append({"role": self.user_prefix, "content": text})
        chat_history_str: str = format_chat_history(chat_history_copy)
        params.metadata.turn_id += 1
        if not params.memory.function_calling_trajectory:
            params.memory.function_calling_trajectory = copy.deepcopy(chat_history_copy)
        else:
            params.memory.function_calling_trajectory.extend(chat_history_copy[-2:])

        params.memory.trajectory.append([])

        orch_state: OrchestratorState = OrchestratorState(
            sys_instruct=self.agent_data.prompt,
            bot_config=BotConfig(
                language=self.agent_data.language,
                llm_config=self.llm_config,
            ),
        )
        return text, chat_history_str, params, orch_state

    def check_skip_node(self, node_info: NodeInfo, chat_history_str: str) -> bool:
        if not node_info.attribute.get("can_skipped", False):
            return False

        task = node_info.attribute.get("task", "")
        if not task:
            return False

        prompts = load_prompts(self.agent_data.language)
        prompt = prompts["check_skip_node_prompt"].format(
            chat_history_str=chat_history_str, task=task
        )
        log_context.info(f"prompt for check skip node: {prompt}")

        model_service: ModelService = ModelService(self.llm_config)
        try:
            response_text = model_service.get_response(prompt)
            log_context.info(f"LLM response for task verification: {response_text}")
            response_text = str(response_text).lower().strip()
            return response_text == "yes"
        except Exception as e:
            log_context.error(f"Error in LLM task verification: {str(e)}")
            return False

    def perform_node(
        self,
        orch_state: OrchestratorState,
        node_info: NodeInfo,
        params: OrchestratorParams,
        text: str,
        chat_history_str: str,
        stream_type: StreamType | None,
        message_queue: janus.Queue | None,
    ) -> tuple[NodeResponse, OrchestratorState, OrchestratorParams]:
        # Create initial resource record with common info and output from trajectory
        resource_record: ResourceRecord = ResourceRecord(
            info={
                "resource": node_info.resource,
                "attribute": node_info.attribute,
                "node_id": params.nlugraph.curr_node,
            },
            intent=params.nlugraph.intent,
        )

        # Add resource record to current turn's list
        params.memory.trajectory[-1].append(resource_record)

        # Update orchestrator state
        orch_state.user_message = ConvoMessage(history=chat_history_str, message=text)
        orch_state.function_calling_trajectory = (
            params.memory.function_calling_trajectory
        )
        orch_state.trajectory = params.memory.trajectory
        orch_state.metadata = params.metadata
        orch_state.stream_type = stream_type
        orch_state.message_queue = message_queue
        # Execute the node
        response_state: OrchestratorState
        response_state, node_response = self.executor.step(
            node_info.resource["id"],
            orch_state,
            node_info,
            params.nlugraph.dialog_states,
        )
        # Update params
        params.nlugraph.node_status[node_info.node_id] = node_response.status
        if node_response.slots:
            params.nlugraph.dialog_states = node_response.slots
        return node_info, response_state, params, node_response

    def execute(
        self,
        inputs: dict[str, Any],
        stream_type: StreamType | None,
        message_queue: janus.Queue | None,
    ) -> tuple[str, str, OrchestratorParams, OrchestratorState]:
        text, chat_history_str, params, orch_state = self.init_params(inputs)
        # NLU Graph Chain
        nlugraph_inputs: dict[str, Any] = {
            "text": text,
            "chat_history_str": chat_history_str,
            "nlu_params": params.nlugraph,
            "allow_global_intent_switch": True,
        }

        orch_state.trajectory = params.memory.trajectory
        nlugraph_chain = RunnableLambda(self.nlu_graph.get_node)

        n_node_performed = 0
        max_n_node_performed = 5
        while n_node_performed < max_n_node_performed:
            nlugraph_start_time = time.time()
            node_info, nlu_params = nlugraph_chain.invoke(nlugraph_inputs)
            nlugraph_inputs["allow_global_intent_switch"] = False
            params.nlugraph = nlu_params
            params.metadata.timing.nlugraph = time.time() - nlugraph_start_time
            # Check if current node can be skipped
            can_skip = self.check_skip_node(node_info, chat_history_str)
            if can_skip:
                continue
            log_context.info(f"The current node info is : {node_info}")

            # perform node
            node_info, orch_state, params, node_response = self.perform_node(
                orch_state,
                node_info,
                params,
                text,
                chat_history_str,
                stream_type,
                message_queue,
            )

            n_node_performed += 1
            # If the current node is not complete, then no need to continue to the next node
            if node_response.status == StatusEnum.INCOMPLETE:
                break
            # If the current node has a response, break the loop
            if node_response.response:
                break
            # If the current node is a leaf node, break the loop
            if node_info.is_leaf is True:
                break

        if not node_response.response:
            log_context.info("No response, do context generation")
            if stream_type == StreamType.NON_STREAM:
                answer = ToolGenerator.context_generate(orch_state)
                node_response.response = answer
            else:
                answer = ToolGenerator.stream_context_generate(orch_state)
                node_response.response = answer
        node_response = post_process_response(
            orch_state,
            node_response,
        )
        return node_response, params



"""Slot filling implementation for NLU.

This module provides the core implementation for slot filling functionality,
supporting both local model-based and remote API-based approaches. It implements
the BaseSlotFilling interface to provide a unified way of extracting and
verifying slot values from input text.

The module includes:
- SlotFiller: Main class for slot filling
- Support for both local and remote slot filling
- Integration with language models and APIs
"""

import json
from typing import Any

from arklex.models.model_service import ModelService
from arklex.orchestrator.nlu.entities.slot_entities import Slot
from arklex.orchestrator.nlu.utils.formatters import (
    format_verification_input as format_verification_input_formatter,
)
from arklex.utils.logging.exceptions import ModelError
from arklex.utils.logging.logging_utils import LogContext, handle_exceptions

log_context = LogContext(__name__)


class SlotFiller:
    """Slot filling implementation.

    This class provides functionality for extracting and verifying slot values
    from user input, supporting both local model-based and remote API-based
    approaches.
    """

    def __init__(
        self,
        model_service: ModelService,
    ) -> None:
        """Initialize the slot filler.

        Args:
            model_service: Service for local model-based slot filling

        Raises:
            ValidationError: If model_service is not provided
        """
        self.model_service = model_service
        log_context.info(
            "SlotFiller initialized successfully",
            extra={
                "mode": "local",
                "operation": "initialization",
            },
        )

    def format_slot_input(
        self, slots: list[dict[str, Any]], context: str, type: str = "chat"
    ) -> tuple[str, str]:
        """Format input for slot filling.

        Creates a prompt for the model to extract slot values from the given context.
        The prompt includes slot definitions and the context to analyze.

        Args:
            slots: List of slot definitions to fill (can be dict or Pydantic model)
            context: Input context to extract values from
            type: Type of slot filling operation (default: "chat")

        Returns:
            Tuple of (user_prompt, system_prompt)
        """
        # Format slot definitions
        slot_definitions = []
        for slot in slots:
            # Handle both dict and Pydantic model inputs
            if isinstance(slot, dict):
                slot_name = slot.get("name", "")
                slot_type = slot.get("type", "string")
                description = slot.get("description", "")
                required = "required" if slot.get("required", False) else "optional"
                items = slot.get("items", {})
            else:
                slot_name = getattr(slot, "name", "")
                slot_type = getattr(slot, "type", "string")
                description = getattr(slot, "description", "")
                required = (
                    "required" if getattr(slot, "required", False) else "optional"
                )
                items = getattr(slot, "items", {})

            slot_def = f"- {slot_name} ({slot_type}, {required}): {description}"
            if items:
                enum_values = (
                    items.get("enum", [])
                    if isinstance(items, dict)
                    else getattr(items, "enum", [])
                )
                if enum_values:
                    slot_def += f"\n  Possible values: {', '.join(enum_values)}"
            slot_definitions.append(slot_def)

        # Create the prompts
        system_prompt = (
            "You are a slot filling assistant. Your task is to extract specific "
            "information from the given context based on the slot definitions. "
            "Extract values for all slots when the information is present in the context, "
            "regardless of whether they are required or optional. "
            "Only set a slot to null if the information is truly not mentioned. "
            "Return the extracted values in JSON format only without any markdown formatting or code blocks."
        )

        user_prompt = (
            f"Context:\n{context}\n\n"
            f"Slot definitions:\n" + "\n".join(slot_definitions) + "\n\n"
            "Please extract the values for the defined slots from the context. "
            "Extract values whenever the information is mentioned, whether the slot is required or optional. "
            "Set to null only if the information is not present in the context. "
            "Return the results in JSON format with slot names as keys and "
            "extracted values as values."
        )

        return user_prompt, system_prompt

    def process_slot_response(
        self, response: str | dict[str, Any], slots: list[dict[str, Any]]
    ) -> list[dict[str, Any]]:
        """Process the model's response for slot filling.

        Parses the model's response and updates the slot values accordingly.
        Handles both traditional slot structures and new slot_schema structures.

        Args:
            response: Model's response containing extracted slot values (can be string or dict)
            slots: Original slot definitions (can be dict or Pydantic model)

        Returns:
            Updated list of slots with extracted values

        Raises:
            ValueError: If response parsing fails
        """
        try:
            # If there are no slots to process, return empty list early
            if not slots:
                return []
            # Handle both string and dict responses
            if isinstance(response, str):
                # Parse the JSON response if it's a string
                extracted_values = json.loads(response)
            elif isinstance(response, dict):
                # Use the dict directly if it's already a dict
                extracted_values = response
            else:
                raise ValueError(f"Unsupported response type: {type(response)}")

            # Route to slot_schema handler if ANY slot provides a slot_schema
            has_any_slot_schema = any(getattr(s, "slot_schema", None) for s in slots)
            if has_any_slot_schema:
                # Handle new slot_schema structure
                return self._process_slot_schema_response(extracted_values, slots)
            else:
                # Handle traditional slot structure
                return self._process_traditional_slot_response(extracted_values, slots)

        except json.JSONDecodeError as e:
            log_context.error(f"Error parsing slot filling response: {str(e)}")
            raise ValueError(f"Failed to parse slot filling response: {str(e)}") from e
        except Exception as e:
            log_context.error(f"Error processing slot filling response: {str(e)}")
            raise ValueError(
                f"Failed to process slot filling response: {str(e)}"
            ) from e

    def _process_slot_schema_response(
        self, extracted_values: dict, slots: list
    ) -> list:
        """Process response for slot_schema structure.

        Args:
            extracted_values: Extracted values from model response
            slots: Original slot definitions

        Returns:
            Updated list of slots with extracted values
        """

        if isinstance(extracted_values, dict) and len(slots) >= 1:
            for s in slots:
                slot_name = (
                    s.get("name") if isinstance(s, dict) else getattr(s, "name", None)
                )
                if not slot_name:
                    continue
                value = extracted_values.get(slot_name)
                if value is not None:
                    if isinstance(s, dict):
                        s["value"] = value
                    else:
                        s.value = value
            return slots

        return slots

    def _process_traditional_slot_response(
        self, extracted_values: dict, slots: list
    ) -> list:
        """Process response for traditional slot structure.

        Args:
            extracted_values: Extracted values from model response
            slots: Original slot definitions

        Returns:
            Updated list of slots with extracted values
        """
        # Update slot values
        for slot in slots:
            # Handle both dict and Pydantic model inputs
            if isinstance(slot, dict):
                slot_name = slot.get("name", "")
                slot["value"] = extracted_values.get(slot_name)
            else:
                slot_name = getattr(slot, "name", "")
                slot.value = extracted_values.get(slot_name)

        return slots

    def format_verification_input(
        self, slot: dict[str, Any], chat_history_str: str
    ) -> str:
        """Format input for slot verification.

        Creates a prompt for the model to verify if a slot value is correct and valid.

        Args:
            slot: Slot definition with value to verify
            chat_history_str: Chat history context

        Returns:
            str: Formatted verification prompt
        """
        return format_verification_input_formatter(slot, chat_history_str)

    def process_verification_response(self, response: str) -> tuple[bool, str]:
        """Process the model's response for slot verification.

        Parses the model's response to determine if verification is needed.

        Args:
            response: Model's response for verification

        Returns:
            Tuple[bool, str]: (verification_needed, reason)
        """
        try:
            # Parse JSON response from formatters
            log_context.info(f"Verification response: {response}")
            response_data = json.loads(response)
            verification_needed = response_data.get("verification_needed", True)
            thought = response_data.get("thought", "No reasoning progivided")
            return verification_needed, thought
        except json.JSONDecodeError as e:
            log_context.error(f"Error parsing verification response: {str(e)}")
            # Default to needing verification if JSON parsing fails
            return True, f"Failed to parse verification response: {str(e)}"

    def format_slot_schema_input(
        self, function_def: dict[str, Any], context: str
    ) -> tuple[str, str]:
        """Format input for slot extraction using an OpenAI-style function schema.

        Args:
            function_def: Dict with keys 'name','description','parameters' matching OpenAI function schema
            context: Input context to extract values from

        Returns:
            Tuple of (user_prompt, system_prompt)
        """
        name = function_def.get("name", "tool")
        description = function_def.get("description", "")
        parameters = function_def.get("parameters", {})
        # Provide the parameters JSON exactly to the model
        schema_json = json.dumps(parameters, ensure_ascii=False)

        system_prompt = (
            "You are a precise information extraction assistant. "
            "Given a JSON Schema for function parameters and a conversation context, "
            "extract values strictly matching the schema. Return ONLY a JSON object that conforms to the 'parameters' schema. "
            "Do not include Markdown or explanations."
        )
        user_prompt = (
            f"Function: {name}\n"
            f"Description: {description}\n"
            f"Parameters JSON Schema:\n{schema_json}\n\n"
            f"Context:\n{context}\n\n"
            "Return a JSON object whose keys and value types exactly match the 'properties' under the parameters schema."
        )
        return user_prompt, system_prompt

    def _slots_to_openai_schema(self, slots: list[Slot]) -> dict[str, Any]:
        """Convert a list of Slot objects into a single OpenAI JSON schema.

        The returned schema is always a top-level object where each slot is a
        property whose value is that slot's own schema definition.

        Preference order per slot:
        1) slot.slot_schema (deep-copied and sanitized), else
        2) slot.to_openai_schema().
        """
        import copy

        properties: dict[str, Any] = {}
        required: list[str] = []

        for slot in slots:
            if getattr(slot, "valueSource", None) == "fixed":
                continue
            # Prefer explicit slot_schema if provided
            if getattr(slot, "slot_schema", None):
                slot_schema = copy.deepcopy(slot.slot_schema)
                # Unwrap legacy function wrapper to plain JSON Schema parameters if present
                if isinstance(slot_schema, dict) and "function" in slot_schema:
                    params = slot_schema.get("function", {}).get("parameters")
                    if isinstance(params, dict):
                        # Extract the specific property's schema for this slot
                        props = (
                            params.get("properties", {})
                            if isinstance(params.get("properties"), dict)
                            else {}
                        )
                        if slot.name in props and isinstance(
                            props.get(slot.name), dict
                        ):
                            slot_schema = props.get(slot.name)
                        else:
                            # Fallback to the parameters block (best-effort) if property missing
                            slot_schema = params
                # Sanitize custom fields to adhere to OpenAI JSON schema
                self._remove_non_openai_fields(slot_schema)
            else:
                # Fall back to the legacy per-slot schema generator
                slot_schema = slot.to_openai_schema()

            if slot_schema is None:
                # Skip slots that explicitly indicate no schema (e.g., fixed values)
                continue

            properties[slot.name] = slot_schema

            if getattr(slot, "required", False):
                required.append(slot.name)

        return {
            "title": "SlotFillingOutput",
            "description": "Structured output for slot filling",
            "type": "object",
            "properties": properties,
            "required": required,
        }

    def _remove_non_openai_fields(self, schema_obj: dict | list) -> None:
        """Recursively remove non-OpenAI standard fields from schema objects.

        Args:
            schema_obj: The schema object to clean (dict or list)
        """
        if isinstance(schema_obj, dict):
            # Remove non-OpenAI fields
            fields_to_remove = [
                "valueSource",
                "fixed",
                "default",
                "prompt",
                "verified",
                "required",
                "repeatable",
            ]
            for field in fields_to_remove:
                schema_obj.pop(field, None)

            # Recursively clean nested objects
            for _key, value in list(schema_obj.items()):
                if isinstance(value, dict | list):
                    self._remove_non_openai_fields(value)
        elif isinstance(schema_obj, list):
            # Recursively clean list items
            for item in schema_obj:
                if isinstance(item, dict | list):
                    self._remove_non_openai_fields(item)

    def _apply_fixed_or_default_for_simple(
        self, slot: Slot, schema_props: dict[str, Any]
    ) -> None:
        """Apply fixed or default value for a simple (non-array) slot using schema properties.
        This preserves current behavior: fixed overrides; default fills only when empty; marks provenance and verified.
        """
        try:
            field_schema = schema_props.get(slot.name)
            if not isinstance(field_schema, dict):
                return
            # Fixed overrides model output
            if field_schema.get("valueSource") == "fixed" and "value" in field_schema:
                slot.value = self._convert_value_to_type(
                    field_schema["value"], field_schema.get("type", "string")
                )
                try:
                    slot.valueSource = "fixed"
                    slot.verified = True
                except Exception:
                    pass
                return

            # Default applies only if no model value
            default_value = None
            if "default" in field_schema:
                default_value = field_schema["default"]
            elif (
                field_schema.get("valueSource") == "default" and "value" in field_schema
            ):
                default_value = field_schema["value"]

            if default_value is not None and (slot.value in (None, "", [])):
                slot.value = self._convert_value_to_type(
                    default_value, field_schema.get("type", "string")
                )
                try:
                    slot.valueSource = "default"
                    slot.verified = True
                except Exception:
                    pass
        except Exception as e:
            log_context.warning(
                f"Failed to apply schema defaults for simple slot {slot.name}: {e}",
                extra={"operation": "slot_filling_evaluation"},
            )

    @handle_exceptions()
    def _fill_slots(
        self,
        slots: list[Slot],
        context: str,
        model_config: dict[str, Any],
        type: str = "chat",
    ) -> list[Slot]:
        """Fill slots.

        Args:
            slots: List of slots to fill
            context: Input context to extract values from
            model_config: Model configuration
            type: Type of slot filling operation (default: "chat")

        Returns:
            List of filled slots

        Raises:
            ModelError: If slot filling fails
            ValidationError: If input validation fails
        """
        # Format input
        prompt, system_prompt = self.format_slot_input(slots, context, type)
        log_context.info(
            "Slot filling input prepared",
            extra={
                "prompt": prompt,
                "system_prompt": system_prompt,
                "operation": "slot_filling_local",
            },
        )

        # Generate OpenAI schema from slots (let LLM handle descriptions/prompts)
        schema = self._slots_to_openai_schema(slots)
        log_context.info(
            "OpenAI schema generated",
            extra={
                "schema": schema,
                "operation": "slot_filling_local",
            },
        )

        # If there are no properties to extract (all slots are fixed or none require LLM),
        # skip calling the model and only apply fixed values. Defaults must be sent to LLM first.
        if not schema.get("properties"):
            log_context.info(
                "No variable slots to extract; skipping LLM call and applying fixed values only",
                extra={
                    "operation": "slot_filling_local",
                },
            )

            # Start from the input slots and fill fixed values for simple slots
            filled_slots = list(slots)
            simple_fixed_slots = [
                s
                for s in slots
                if not getattr(s, "slot_schema", None)
                and getattr(s, "valueSource", None) == "fixed"
            ]
            if simple_fixed_slots:
                name_to_slot = {s.name: s for s in filled_slots}
                for fixed_slot in simple_fixed_slots:
                    if fixed_slot.name in name_to_slot:
                        fixed_value = getattr(fixed_slot, "fixed", None)
                        if fixed_value is not None:
                            target = name_to_slot[fixed_slot.name]
                            target.value = fixed_value
                            try:
                                target.valueSource = "fixed"
                                target.verified = True
                            except Exception:
                                pass

            log_context.info(
                "Slot filling completed without LLM (fixed only)",
                extra={
                    "filled_slots": [slot.name for slot in filled_slots],
                    "operation": "slot_filling_local",
                },
            )
            return filled_slots

        # Get model response (guard against None)
        response = self.model_service.get_response_with_structured_output(
            prompt, schema, system_prompt
        )
        log_context.info(
            "Model response received",
            extra={
                "prompt": prompt,
                "system_prompt": system_prompt,
                "raw_response": response,
                "operation": "slot_filling_local",
            },
        )

        # Process response
        try:
            # If model produced no structured output, skip processing and just apply fixed/defaults
            if response is None:
                log_context.warning(
                    "Model returned None; skipping response processing and applying fixed/default values",
                    extra={
                        "operation": "slot_filling_local",
                    },
                )
                filled_slots = list(slots)
            else:
                filled_slots = self.process_slot_response(response, slots)

            # Apply default/fixed values for any slots that use slot_schema
            if slots:
                schema_slots = [s for s in slots if getattr(s, "slot_schema", None)]
                for schema_slot in schema_slots:
                    filled_slots = self._evaluate_and_fill_slot_values(
                        filled_slots, schema_slot
                    )
                # Also apply fixed/default values for simple, non-nested slots that were skipped
                simple_fixed_slots = [
                    s
                    for s in slots
                    if not getattr(s, "slot_schema", None)
                    and getattr(s, "valueSource", None) == "fixed"
                ]
                if simple_fixed_slots:
                    name_to_slot = {s.name: s for s in filled_slots}
                    for fixed_slot in simple_fixed_slots:
                        if fixed_slot.name in name_to_slot:
                            # Prefer explicit fixed value, else default
                            fixed_value = getattr(fixed_slot, "fixed", None)
                            if fixed_value is None:
                                fixed_value = getattr(fixed_slot, "default", None)
                            if fixed_value is not None:
                                target = name_to_slot[fixed_slot.name]
                                target.value = fixed_value
                                # Mark provenance to avoid verification later
                                try:
                                    target.valueSource = "fixed"
                                    target.verified = True
                                except Exception:
                                    pass

                # Apply defaults for simple non-fixed slots if model didn't provide values
                defaultable_slots = [
                    s
                    for s in slots
                    if getattr(s, "valueSource", None) != "fixed"
                    and getattr(s, "default", None) is not None
                ]
                if defaultable_slots:
                    name_to_slot = {s.name: s for s in filled_slots}
                    for default_slot in defaultable_slots:
                        if default_slot.name in name_to_slot:
                            target = name_to_slot[default_slot.name]
                            if getattr(target, "value", None) in (None, "", []):
                                target.value = default_slot.default
                                # Mark provenance for defaults; verification not needed
                                try:
                                    target.valueSource = "default"
                                    target.verified = True
                                except Exception:
                                    pass

            log_context.info(
                "Slot filling completed",
                extra={
                    "prompt": prompt,
                    "system_prompt": system_prompt,
                    "raw_response": response,
                    "filled_slots": [slot.name for slot in filled_slots],
                    "operation": "slot_filling_local",
                },
            )
            return filled_slots
        except Exception as e:
            log_context.error(
                "Failed to process slot filling response",
                extra={
                    "prompt": prompt,
                    "system_prompt": system_prompt,
                    "raw_response": response,
                    "error": str(e),
                    "operation": "slot_filling_local",
                },
            )
            raise ModelError(
                "Failed to process slot filling response",
                details={
                    "prompt": prompt,
                    "system_prompt": system_prompt,
                    "raw_response": response,
                    "error": str(e),
                    "operation": "slot_filling_local",
                },
            ) from e

    def _evaluate_and_fill_slot_values(
        self, filled_slots: list[Slot], original_slot: Slot
    ) -> list[Slot]:
        """Evaluate and fill back default and fixed values from slot_schema structure.

        Args:
            filled_slots: List of slots with model-extracted values
            original_slot: Original slot with slot_schema structure

        Returns:
            Updated list of slots with proper values filled
        """
        if not original_slot.slot_schema:
            return filled_slots

        # For the new slot_schema structure, we need to handle the nested array structure
        for slot in filled_slots:
            if (
                slot.name == original_slot.name
                and slot.value
                and isinstance(slot.value, list)
            ):
                # This is an array slot, we need to process each item
                updated_items = []
                for item in slot.value:
                    if isinstance(item, dict):
                        # Apply fixed values to each item in the array using direct field access
                        updated_item = self._apply_fixed_values_direct(
                            item, original_slot.slot_schema
                        )
                        updated_items.append(updated_item)
                    else:
                        updated_items.append(item)
                slot.value = updated_items

                log_context.info(
                    f"Applied fixed values to array slot {slot.name}",
                    extra={
                        "slot_name": slot.name,
                        "updated_value": slot.value,
                        "operation": "slot_filling_evaluation",
                    },
                )
            elif slot.name == original_slot.name and (slot.value in (None, "", [])):
                # Simple (non-array) slot: apply fixed/default from schema via helper
                props = (
                    original_slot.slot_schema.get("function", {})
                    .get("parameters", {})
                    .get("properties", {})
                )
                self._apply_fixed_or_default_for_simple(slot, props)

        return filled_slots

    def _apply_fixed_values_direct(self, item: dict, slot_schema: dict) -> dict:
        """Apply fixed values directly to an item using field-level access.

        Args:
            item: Dictionary item to update
            slot_schema: The slot schema containing field definitions

        Returns:
            Updated item with fixed values applied
        """
        try:
            # Get the array items schema directly
            slot_name = None
            for key in (
                slot_schema.get("function", {})
                .get("parameters", {})
                .get("properties", {})
            ):
                slot_name = key
                break

            if not slot_name:
                return item

            # Get the array items schema
            array_schema = (
                slot_schema.get("function", {})
                .get("parameters", {})
                .get("properties", {})
                .get(slot_name, {})
            )
            items_schema = array_schema.get("items", {})
            properties = items_schema.get("properties", {})

            # Apply fixed values recursively to the item
            updated_item = self._apply_fixed_values_recursive(item, properties)

            return updated_item

        except Exception as e:
            log_context.error(
                "Error applying fixed values to item",
                extra={
                    "error": str(e),
                    "item": item,
                    "operation": "slot_filling_evaluation",
                },
            )
            return item

    def _apply_fixed_values_recursive(
        self, item: dict, properties: dict, path: str = ""
    ) -> dict:
        """Recursively apply fixed values to an item and its nested structures.

        Args:
            item: Dictionary item to update
            properties: Properties schema containing field definitions
            path: Current path for nested fields

        Returns:
            Updated item with fixed values applied
        """
        updated_item = item.copy()

        for field_name, field_schema in properties.items():
            current_path = f"{path}.{field_name}" if path else field_name

            # Check if this field has a fixed or default
            if field_schema.get("valueSource") == "fixed" and "value" in field_schema:
                # Convert and apply the fixed value
                fixed_value = self._convert_value_to_type(
                    field_schema["value"], field_schema.get("type", "string")
                )
                updated_item[field_name] = fixed_value
                log_context.info(
                    f"Applied fixed value to field {current_path}",
                    extra={
                        "field_path": current_path,
                        "fixed_value": fixed_value,
                        "original_value": field_schema["value"],
                        "type": field_schema.get("type", "string"),
                        "operation": "slot_filling_evaluation",
                    },
                )

            elif "default" in field_schema:
                # Default only applies if the model didn't provide a value
                if field_name not in updated_item or updated_item[field_name] in (
                    None,
                    "",
                    [],
                ):
                    default_value = self._convert_value_to_type(
                        field_schema["default"], field_schema.get("type", "string")
                    )
                    updated_item[field_name] = default_value
                    log_context.info(
                        f"Applied default value to field {current_path}",
                        extra={
                            "field_path": current_path,
                            "default_value": default_value,
                            "operation": "slot_filling_evaluation",
                        },
                    )

            # Handle nested objects
            elif field_schema.get("type") == "object" and field_name in updated_item:
                nested_props = field_schema.get("properties", {})
                if nested_props and isinstance(updated_item[field_name], dict):
                    updated_item[field_name] = self._apply_fixed_values_recursive(
                        updated_item[field_name], nested_props, current_path
                    )

            # Handle arrays of objects
            elif field_schema.get("type") == "array" and field_name in updated_item:
                items_schema = field_schema.get("items", {})
                if items_schema.get("type") == "object" and isinstance(
                    updated_item[field_name], list
                ):
                    nested_props = items_schema.get("properties", {})
                    if nested_props:
                        updated_item[field_name] = [
                            self._apply_fixed_values_recursive(
                                item, nested_props, current_path
                            )
                            for item in updated_item[field_name]
                        ]

        return updated_item

    def _convert_value_to_type(
        self, value: str | int | float | bool | list | dict | None, target_type: str
    ) -> str | int | float | bool | list | dict | None:
        """Convert a value to the specified type.

        Args:
            value: The value to convert
            target_type: The target type string

        Returns:
            The converted value
        """
        try:
            if target_type == "boolean":
                if isinstance(value, str):
                    return value.lower() in ("true", "1", "yes", "on")
                return bool(value)
            elif target_type == "integer":
                return int(value)
            elif target_type == "number":
                return float(value)
            else:  # string or unknown type
                return str(value)
        except (ValueError, TypeError) as e:
            log_context.warning(
                f"Failed to convert value {value} to type {target_type}",
                extra={
                    "value": value,
                    "target_type": target_type,
                    "error": str(e),
                    "operation": "slot_filling_type_conversion",
                },
            )
            return value

    @handle_exceptions()
    def _verify_slot_local(
        self,
        slot: dict[str, Any],
        chat_history_str: str,
        model_config: dict[str, Any],
    ) -> tuple[bool, str]:
        """Verify slot value using local model.

        Args:
            slot: Slot to verify
            chat_history_str: Formatted chat history
            model_config: Model configuration

        Returns:
            Tuple of (is_valid, reason)

        Raises:
            ModelError: If slot verification fails
            ValidationError: If input validation fails
        """
        log_context.info(
            "Using local model for slot verification",
            extra={
                "slot": slot.get("name", "unknown"),
                "operation": "slot_verification_local",
            },
        )

        # Format input
        prompt = self.format_verification_input(slot, chat_history_str)
        log_context.info(
            "Slot verification input prepared",
            extra={
                "prompt": prompt,
                "operation": "slot_verification_local",
            },
        )

        # Get model response
        response = self.model_service.get_response(prompt)
        log_context.info(
            "Model response received",
            extra={
                "response": response,
                "operation": "slot_verification_local",
            },
        )

        # Process response
        try:
            is_valid, reason = self.process_verification_response(response)
            log_context.info(
                "Slot verification completed",
                extra={
                    "is_valid": is_valid,
                    "reason": reason,
                    "operation": "slot_verification_local",
                },
            )
            return is_valid, reason
        except Exception as e:
            log_context.error(
                "Failed to process slot verification response",
                extra={
                    "error": str(e),
                    "response": response,
                    "operation": "slot_verification_local",
                },
            )
            raise ModelError(
                "Failed to process slot verification response",
                details={
                    "error": str(e),
                    "response": response,
                    "operation": "slot_verification_local",
                },
            ) from e

    @handle_exceptions()
    def verify_slot(
        self,
        slot: Slot | dict[str, Any],
        chat_history_str: str,
        model_config: dict[str, Any],
    ) -> tuple[bool, str]:
        """Verify slot value using local model.

        Args:
            slot: Slot to verify (can be Slot object or dict)
            chat_history_str: Formatted chat history
            model_config: Model configuration

        Returns:
            Tuple of (is_valid, reason)

        Raises:
            ModelError: If slot verification fails
            ValidationError: If input validation fails
            APIError: If API request fails
        """

        # Short-circuit verification for fixed or default-provisioned slots
        try:
            if hasattr(slot, "valueSource"):
                vs = getattr(slot, "valueSource", None)
                if vs in ("fixed", "default"):
                    return True, f"{vs.capitalize()} value; no verification required"
            if isinstance(slot, dict):
                vs = slot.get("valueSource")
                if vs in ("fixed", "default"):
                    return True, f"{vs.capitalize()} value; no verification required"
        except Exception:
            pass

        # Handle both Slot objects and dictionaries
        slot_name = slot.name if hasattr(slot, "name") else slot.get("name", "unknown")

        # Convert Slot object to dictionary if needed
        slot_dict = slot
        if hasattr(slot, "__dict__"):
            slot_dict = {
                "name": slot.name,
                "value": getattr(slot, "value", None),
                "type": getattr(slot, "type", None),
                "description": getattr(slot, "description", None),
                "enum": getattr(slot, "enum", None),
                "required": getattr(slot, "required", False),
                "repeatable": getattr(slot, "repeatable", False),
                "prompt": getattr(slot, "prompt", None),
                "valueSource": getattr(slot, "valueSource", None),
                "fixed": getattr(slot, "fixed", None),
                "default": getattr(slot, "default", None),
                "verified": getattr(slot, "verified", False),
            }

        log_context.info(
            "Starting slot verification",
            extra={
                "slot": slot_name,
                "mode": "local",
                "operation": "slot_verification",
            },
        )

        try:
            is_valid, reason = self._verify_slot_local(
                slot_dict, chat_history_str, model_config
            )

            log_context.info(
                "Slot verification completed",
                extra={
                    "is_valid": is_valid,
                    "reason": reason,
                    "operation": "slot_verification",
                },
            )
            return is_valid, reason
        except Exception as e:
            log_context.error(
                "Slot verification failed",
                extra={
                    "error": str(e),
                    "slot": slot_name,
                    "operation": "slot_verification",
                },
            )
            raise

    @handle_exceptions()
    def fill_slots(
        self,
        slots: list[Slot],
        context: str,
        model_config: dict[str, Any],
        type: str = "chat",
    ) -> list[Slot]:
        """Fill slots from input context.

        Args:
            slots: List of slots to fill
            context: Input context to extract values from
            model_config: Model configuration
            type: Type of slot filling operation (default: "chat")

        Returns:
            List of filled slots

        Raises:
            ModelError: If slot filling fails
            ValidationError: If input validation fails
            APIError: If API request fails
        """
        log_context.info(
            "Starting slot filling",
            extra={
                "slots": [slot.name for slot in slots],
                "context_length": len(context),
                "mode": "local",
                "operation": "slot_filling",
            },
        )

        try:
            filled_slots = self._fill_slots(slots, context, model_config, type)

            log_context.info(
                "Slot filling completed",
                extra={
                    "filled_slots": [slot.name for slot in filled_slots],
                    "operation": "slot_filling",
                },
            )
            return filled_slots
        except Exception as e:
            log_context.error(
                "Slot filling failed",
                extra={
                    "error": str(e),
                    "slots": [slot.name for slot in slots],
                    "operation": "slot_filling",
                },
            )
            raise



from typing import Any

from arklex.resources.agent_map import AGENT_MAP
from arklex.utils.logging.logging_utils import LogContext

log_context = LogContext(__name__)


class AgentLoader:
    @staticmethod
    def init_agents(agents: list[dict[str, Any]]) -> dict[str, dict[str, Any]]:
        """Initialize agents from configuration.

        Args:
            agents: list of agent configurations

        Returns:
            dictionary mapping agent IDs to their configurations
        """
        agent_registry: dict[str, dict[str, Any]] = {}
        for agent in agents:
            agent_id: str = agent["id"]
            try:
                agent_instance = AGENT_MAP[agent_id]["item_cls"]
                agent_registry[agent_id] = {
                    "agent_instance": agent_instance,
                }
            except Exception as e:
                log_context.error(f"Agent {agent_id} is not registered, error: {e}")
                continue

        return agent_registry
